from dataclasses import dataclass, replace
import random
from scenario import Scenario
from state_update_model import (
    Highlight,
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch8Scenario(Scenario):
    """Challenge Implementation"""

    def __init__(self, seed:int | None = None):
        super().__init__(seed=seed)

    def get_goal_state_description(self) -> str:
        return f"""
            All blue marbles sorted by value in the leftmost column. Lowest value on top.
            All yellow marbles sorted by value in the rightmost column. Lowest value on top.
            Claw 0 can operate in columns 0-2. Claw 1 can operate in columns 4-6.
            \n{self.get_dimensions_description()}
        """

    def get_initial_state(self) -> StateModel:
        max_x = 6
        max_y = 6
        random.seed(self._seed)
        
        value_y_pairs = [(random.randint(0,10), y) for y in range(2,7)]
        blue_balls = [StateBall(pos=StatePosition(x=2, y=y), color="lightblue", value=v, label=f"{v}") for (v,y) in value_y_pairs]
        value_y_pairs = [(random.randint(0,10), y) for y in range(2,7)]
        yellow_balls = [StateBall(pos=StatePosition(x=4, y=y), color="yellow", value=v, label=f"{v}") for (v,y) in value_y_pairs]

        balls = blue_balls + yellow_balls

        claw0 = replace(get_default_state().claws[0], max_x = 2)
        claw1 = replace(claw0, pos=StatePosition(x=6, y=0), min_x=4, max_x = 6)
        claws = [claw0, claw1]

        highlights = [Highlight(xMin=3, xMax=3, yMin=0, yMax=6, color="gray")]

        return replace(
            get_default_state(),
            balls=balls,
            max_x=max_x,
            max_y=max_y,
            claws=claws,
            highlights=highlights,
        )

    def is_in_goal_state(self, state: StateModel) -> bool:
        # No ball in either claw
        if state.claws[0].ball:
            return False
        
        if state.claws[1].ball:
            return False

        column0: list[StateBall] = [ball for ball in state.balls if ball.pos.x == 0]
        column6: list[StateBall] = [ball for ball in state.balls if ball.pos.x == 6]

        if len(column0) != len(state.balls) / 2:
            return False
        
        if len(column6) != len(state.balls) / 2:
            return False
        
        actual_col0_values = [0 if ball.value is None else ball.value for ball in sorted(column0, key=lambda ball: ball.pos.y)]
        expected_col0_values = sorted(actual_col0_values)
        if actual_col0_values != expected_col0_values:
            return False
        
        actual_col6_values = [0 if ball.value is None else ball.value for ball in sorted(column6, key=lambda ball: ball.pos.y)]
        expected_col6_values = sorted(actual_col6_values)
        if actual_col6_values != expected_col6_values:
            return False

        return True



from dataclasses import dataclass, replace
from state_update_model import StateBall, StateModel, StatePosition

@dataclass
class RevealAction:
    """Reveal a ball's value when it's dropped in a specific location."""

    pos: StatePosition

    def on_ball_dropped(self, state: StateModel, ball: StateBall) -> tuple[StateModel, bool]:
        if ball.pos != self.pos:
            return state, False

        ball_positions = [ball.pos for ball in state.balls]
        ball_index = ball_positions.index(self.pos)

        state.balls[ball_index].value_visible = True
        if state.balls[ball_index].label == "?":
            state.balls[ball_index].label = f"{state.balls[ball_index].value}"

        return replace(state, balls=state.balls), True



from dataclasses import dataclass
from typing import ClassVar

MIN_X = 0
MIN_Y = 0

@dataclass
class StatePosition:
    x: int
    y: int

    def __str__(self) -> str:
        return f"x={self.x} y={self.y}"


@dataclass
class StateBall:
    __id_cnt: ClassVar[int] = 0
    pos: StatePosition
    color: str
    value: int | None = None
    label: str = ""    
    value_visible: bool = True
    id: str = "" # always overwritten in __post_init__

    @classmethod
    def __create_unique_id(cls) -> str:
        cls.__id_cnt = cls.__id_cnt + 1
        return f"{cls.__id_cnt}"

    def __post_init__(self):
        if self.id == "":
            self.id = self.__create_unique_id()

    def __str__(self) -> str:
        return f"{self.color} {self.label}"

@dataclass
class Claw:
    pos: StatePosition
    open: bool    
    min_x: int
    max_x: int
    moving_horizontally: bool
    moving_vertically: bool
    operating_claw: bool
    ball: StateBall | None

@dataclass
class Spotlight:
    on: bool
    pos: StatePosition

@dataclass
class Highlight:
    xMin: int
    xMax: int
    yMin: int
    yMax: int
    color: str

@dataclass
class StateModel:
    max_x: int
    max_y: int
    balls: list[StateBall]
    claws: list[Claw]
    goal_accomplished: bool
    spotlight: Spotlight | None
    highlights: list[Highlight] | None
    elapsed: float

@dataclass
class StateUpdateModel:
    userId: str
    state: StateModel
    delay_multiplier: float


def get_default_state() -> StateModel:
    return StateModel(
        max_x=3,
        max_y=4,
        balls=[],
        claws=[Claw(pos=StatePosition(x=0, y=0), open=True, min_x=0, max_x=100, moving_horizontally=False, moving_vertically=False, operating_claw=False, ball=None)],
        goal_accomplished=False,
        spotlight=None,
        highlights=None,
        elapsed=0.0
    )



from dataclasses import dataclass, replace
from scenario import Scenario
from state_update_model import (
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch4Scenario(Scenario):
    """Challenge Implementation"""

    def get_goal_state_description(self) -> str:
        return f"Tower of Hanoi\nMove all marbles from the leftmost column to the rightmost column. A marble can not be placed on top of a marble of lower value.\n{self.get_dimensions_description()}"
    
    def get_initial_state(self) -> StateModel:
        max_x = 2
        max_y = 5
        balls = [StateBall(pos=StatePosition(x=0, y=y), color="lightgreen", value=y+1, label=f"{y+1}") for y in range(max_y + 1)]
        
        return replace(get_default_state(), balls = balls, max_x=max_x, max_y=max_y)

    def is_in_goal_state(self, state: StateModel) -> bool:

        # No ball in claw
        if state.claws[0].ball:
            return False
        
        def validate_ball(ball: StateBall) -> bool:
            # ball must be in rightmost column
            if ball.pos.x != state.max_x:
                return False
            return True

        return not next((True for ball in state.balls if not validate_ball(ball)), False)
    
    


from dataclasses import dataclass
from ch4_state_validator import Ch4StateValidator
from scenario import Scenario

from state_manager import StateManager

@dataclass
class Ch4StateManager(StateManager):
    """Validates operations and keeps state up to date"""

    def __init__(self, scenario : Scenario | None = None):
        super().__init__(scenario=scenario)
        self.validator = Ch4StateValidator()



import asyncio
import sys

sys.path.append("../src/ballsort")

from ball_control import BallControl
from state_update_model import StateBall, StatePosition

async def noop():
    pass

async def go_to_pos(bc: BallControl, dest: StatePosition, open_claw: bool, claw_index: int = 0):
    rel_x = dest.x - bc.get_position(claw_index=claw_index).x
    rel_y = dest.y - bc.get_position(claw_index=claw_index).y
    await asyncio.gather(
        bc.move_horizontally(rel_x, claw_index=claw_index),
        bc.move_vertically(rel_y, claw_index=claw_index),
        bc.open_claw(claw_index=claw_index) if open_claw else noop(),
    )

async def move_ball(
    bc: BallControl, src: StatePosition, dest: StatePosition, claw_index: int = 0
):
    await go_to_pos(bc=bc, dest=src, open_claw=True, claw_index=claw_index)
    await bc.close_claw(claw_index=claw_index)
    await go_to_pos(bc=bc, dest=dest, open_claw=False, claw_index=claw_index)
    await bc.open_claw(claw_index=claw_index)

def get_column_top_occupied_y(bc: BallControl, x: int) -> int:
    return min(
        [ball.pos.y for ball in bc.get_state().balls if ball.pos.x == x],
        default=bc.get_state().max_y,
    )

def get_column_top_occupied_pos(bc: BallControl, x: int) -> StatePosition:
    return StatePosition(x=x, y=get_column_top_occupied_y(bc=bc, x=x))

def get_column_top_vacant_y(bc: BallControl, x: int) -> int:
    return (
        min(
            [ball.pos.y for ball in bc.get_state().balls if ball.pos.x == x],
            default=bc.get_state().max_y + 1,
        )
        - 1
    )

def get_column_top_vacant_pos(bc: BallControl, x: int) -> StatePosition:
    return StatePosition(x=x, y=get_column_top_vacant_y(bc=bc, x=x))

async def move_ball_by_column(
    bc: BallControl, src_x: int, dest_x: int, claw_index: int = 0
):
    await move_ball(
        bc=bc,
        src=get_column_top_occupied_pos(bc=bc, x=src_x),
        dest=get_column_top_vacant_pos(bc=bc, x=dest_x),
        claw_index=claw_index,
    )


async def sort_column(
    bc: BallControl,
    src_x1: int,
    src_x2: int,
    dest_x: int,
    nof_balls: int,
    claw_index: int,
):
    """takes balls from columns src_x1 and src_x2 and puts them, ordered by value, in column dest_x"""

    for _ in range(nof_balls):
        column1: list[StateBall] = [
            ball for ball in bc.get_state().balls if ball.pos.x == src_x1
        ]
        column2: list[StateBall] = [
            ball for ball in bc.get_state().balls if ball.pos.x == src_x2
        ]
        column1_sorted = [
            0 if ball.value is None else ball.value
            for ball in sorted(column1, key=lambda ball: ball.pos.y)
        ]
        column2_sorted = [
            0 if ball.value is None else ball.value
            for ball in sorted(column2, key=lambda ball: ball.pos.y)
        ]

        if max(column1_sorted, default=-1000) >= max(column2_sorted, default=-1000):
            src_column_index = src_x1
            dest_column_index = src_x2
            src_column = column1_sorted
        else:
            src_column_index = src_x2
            dest_column_index = src_x1
            src_column = column2_sorted

        minpos = src_column.index(max(src_column))

        for _ in range(minpos):
            await move_ball_by_column(
                bc=bc,
                src_x=src_column_index,
                dest_x=dest_column_index,
                claw_index=claw_index,
            )

        await move_ball_by_column(
            bc=bc, src_x=src_column_index, dest_x=dest_x, claw_index=claw_index
        )



from dataclasses import dataclass, replace
from scenario import Scenario
from state_update_model import (
    Highlight,
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch6Scenario(Scenario):
    """Challenge Implementation"""

    def get_goal_state_description(self) -> str:
        return f"Blue in leftmost column. Yellow in rightmost column. Left claw can operate in columns 0-2. Right claw can operate in columns 2-4.\n{self.get_dimensions_description()}"
    
    def get_initial_state(self) -> StateModel:
        max_x = 4
        max_y = 4
        balls = [
            StateBall(pos=StatePosition(x=0, y=4), color="yellow"),
            StateBall(pos=StatePosition(x=4, y=4), color="blue"),
        ]


        claw0 = replace(get_default_state().claws[0], max_x = 2)
        claw1 = replace(claw0, pos=StatePosition(x=4, y=0), min_x=2, max_x = 4)
        claws = [claw0, claw1]

        highlights = [Highlight(xMin=2, xMax=2, yMin=0, yMax=4, color="#eeeeee")]
        
        return replace(get_default_state(), balls = balls, max_x=max_x, max_y=max_y, claws=claws, highlights=highlights)

    def is_in_goal_state(self, state: StateModel) -> bool:

        # No ball in either claw
        if state.claws[0].ball:
            return False
        
        if state.claws[1].ball:
            return False
        
        column0: list[StateBall] = [ball for ball in state.balls if ball.pos.x == 0]
        column4: list[StateBall] = [ball for ball in state.balls if ball.pos.x == 4]

        if len(column0) != 1 or column0[0].color != "blue":
            return False
        
        if len(column4) != 1 or column4[0].color != "yellow":
            return False
        
        return True
    


from dataclasses import dataclass, replace
import random
from reveal_action import RevealAction
from scenario import Scenario
from state_update_model import (
    Highlight,
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch11Scenario(Scenario):
    """Challenge Implementation"""

    max_x = 8
    max_y = 6
    left_color = "lightblue"
    right_color = "yellow"
    reveal_action = RevealAction(pos=StatePosition(x=max_x // 2, y=max_y))

    def __init__(self, seed: int | None = None):
        super().__init__(seed=seed)

    def get_goal_state_description(self) -> str:
        return f"""
Blue marbles sorted by value in the leftmost column. Lowest value on top.
Yellow marbles sorted by value in the rightmost column.
Each marble has an, initially hidden, integer value in the range [1, {(self.max_x//2)-1}].
A marble must be dropped on position ({self.max_x//2}, {self.max_y}) to reveal its value.
Claw 0 can operate in columns 0-{self.max_x//2}. Claw 1 can operate in columns {self.max_x//2}-{self.max_x}.
{self.get_dimensions_description()}"""

    def get_initial_state(self) -> StateModel:
        random.seed(self._seed)

        def get_value_y_pairs():
            return [(random.randint(1, (self.max_x // 2) - 1), y) for y in range(1, 7)]

        blue_balls = [
            StateBall(
                pos=StatePosition(x=0, y=y),
                color=self.left_color,
                value=v,
                label="?",
                value_visible=False,
            )
            for (v, y) in get_value_y_pairs()
        ]
        yellow_balls = [
            StateBall(
                pos=StatePosition(x=self.max_x, y=y),
                color=self.right_color,
                value=v,
                label="?",
                value_visible=False,
            )
            for (v, y) in get_value_y_pairs()
        ]

        highlights = [
            Highlight(
                xMin=self.max_x // 2,
                xMax=self.max_x // 2,
                yMin=self.max_y,
                yMax=self.max_y,
                color="lightyellow",
            )
        ]

        claw0 = replace(get_default_state().claws[0], max_x = self.max_x//2)
        claw1 = replace(claw0, pos=StatePosition(x=6, y=0), min_x=self.max_x//2, max_x = self.max_x)

        return replace(
            get_default_state(),
            balls=blue_balls + yellow_balls,
            max_x=self.max_x,
            max_y=self.max_y,
            highlights=highlights,
            claws=[claw0, claw1]
        )

    def _is_column_in_goal_state(self, state: StateModel, x: int, color: str) -> bool:
        column: list[StateBall] = [
            ball for ball in state.balls if ball.pos.x == x and ball.color == color
        ]

        if len(column) != len(state.balls) // 2:
            return False

        actual_values = [
            0 if ball.value is None else ball.value
            for ball in sorted(column, key=lambda ball: ball.pos.y)
        ]
        expected_values = sorted(actual_values)

        return expected_values == actual_values

    def is_in_goal_state(self, state: StateModel) -> bool:
        # No ball in claw
        if state.claws[0].ball:
            return False

        return self._is_column_in_goal_state(
            state=state, x=0, color=self.left_color
        ) and self._is_column_in_goal_state(
            state=state, x=self.max_x, color=self.right_color
        )

    def on_ball_dropped(
        self, state: StateModel, ball: StateBall
    ) -> tuple[StateModel, bool]:
        """Override"""
        return self.reveal_action.on_ball_dropped(state=state, ball=ball)



from dataclasses import dataclass, replace
import random
from scenario import Scenario
from state_update_model import (
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch7Scenario(Scenario):
    """Challenge Implementation"""

    def __init__(self, seed:int | None = None):
        super().__init__(seed=seed)

    def get_goal_state_description(self) -> str:
        return f"All marbles sorted by value in the leftmost column. Lowest value on top."
    
    def get_initial_state(self) -> StateModel:
        max_x = 4
        max_y = 6
        random.seed(self._seed)

        value_y_pairs = [(random.randint(0,10), y) for y in range(2,7)]
        balls = [StateBall(pos=StatePosition(x=1, y=y), color="yellow", value=v, label=f"{v}") for (v,y) in value_y_pairs]

        return replace(get_default_state(), balls = balls, max_x=max_x, max_y=max_y)

    def is_in_goal_state(self, state: StateModel) -> bool:

        # No ball in claw
        if state.claws[0].ball:
            return False
        
        column0: list[StateBall] = [ball for ball in state.balls if ball.pos.x == 0]

        if len(column0) != len(state.balls):
            return False
        
        actual_values = [0 if ball.value is None else ball.value for ball in sorted(state.balls, key=lambda ball: ball.pos.y)]
        expected_values = sorted(actual_values)

        #print(expected_values, actual_values)

        return expected_values == actual_values
    


from dataclasses import dataclass, replace
from scenario import Scenario
from state_update_model import (
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch5Scenario(Scenario):
    """Challenge Implementation"""

    def get_goal_state_description(self) -> str:
        return f"Bicolor Towers of Hanoi\nMove all marbles to two separate columns, one of each color. A marble can not be placed on top of a marble of lower value.\n{self.get_dimensions_description()}"
    
    def get_initial_state(self) -> StateModel:
        balls: list[StateBall] = []
        max_x = 3
        max_y = 5
        for y in range(3):
            value=y+1
            label=f"{y+1}"
            balls.append(StateBall(pos=StatePosition(x=0, y=2*y), color="lightblue", value=value, label=label))
            balls.append(StateBall(pos=StatePosition(x=0, y=2*y+1), color="lightgreen", value=value, label=label))
        
        return replace(get_default_state(), balls = balls, max_x=max_x, max_y=max_y)

    def is_in_goal_state(self, state: StateModel) -> bool:

        # No ball in claw
        if state.claws[0].ball:
            return False
        
        columns: list[list[StateBall]] = [[] for _ in range(state.max_x + 1)]
        for ball in state.balls:
            columns[ball.pos.x].append(ball)

        def column_is_goal_state_compliant(x: int) -> bool:
            column = columns[x]
            if len(column) == 0:
                return True # empty column is ok
            
            bottom_ball = column[0]
            # Column must contain no ball with color different than the bottom ball color
            return not next((True for ball in column if ball.color != bottom_ball.color), False)
        
        # There shall be exactly two populated columns
        if len([True for column in columns if len(column) > 0]) != 2:
            return False

        # There must be no non-compliant columns
        return not next((True for x in range(state.max_x + 1) if not column_is_goal_state_compliant(x)), False)
    


from dataclasses import dataclass, replace
import random
from reveal_action import RevealAction
from scenario import Scenario
from state_update_model import (
    Highlight,
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch10Scenario(Scenario):
    """Challenge Implementation"""

    max_x = 4
    max_y = 6
    reveal_action = RevealAction(pos=StatePosition(x=max_x, y=max_y))

    def __init__(self, seed:int | None = None):
        super().__init__(seed=seed)

    def get_goal_state_description(self) -> str:
        return f"All marbles sorted by value in the leftmost column. Lowest value on top.\nEach marble has an, initially hidden, integer value in the range [1, 3].\nA marble must be dropped on position ({self.max_x}, {self.max_y}) to reveal its value."
    
    def get_initial_state(self) -> StateModel:
        
        random.seed(self._seed)

        value_y_pairs = [(random.randint(1, 3), y) for y in range(1,7)]
        balls = [StateBall(pos=StatePosition(x=0, y=y), color="cyan", value=v, label="?", value_visible=False) for (v,y) in value_y_pairs]

        highlights = [Highlight(xMin=self.max_x, xMax=self.max_x, yMin=self.max_y, yMax=self.max_y, color="lightyellow")]

        return replace(get_default_state(), balls = balls, max_x=self.max_x, max_y=self.max_y, highlights=highlights)

    def is_in_goal_state(self, state: StateModel) -> bool:

        # No ball in claw
        if state.claws[0].ball:
            return False
        
        column0: list[StateBall] = [ball for ball in state.balls if ball.pos.x == 0]

        if len(column0) != len(state.balls):
            return False
        
        actual_values = [0 if ball.value is None else ball.value for ball in sorted(state.balls, key=lambda ball: ball.pos.y)]
        expected_values = sorted(actual_values)

        return expected_values == actual_values
    
    def on_ball_dropped(self, state: StateModel, ball: StateBall) -> tuple[StateModel, bool]:
        """Override"""
        return self.reveal_action.on_ball_dropped(state=state, ball=ball)
    


from dataclasses import dataclass, replace
from scenario import Scenario
from state_utils import get_ball_at_current_pos
from state_validator import StateValidator
from state_update_model import (
    StateModel,
    StatePosition,
)


@dataclass
class StateManager:
    """Validates operations and keeps state up to date"""

    validator: StateValidator
    scenario: Scenario | None

    def __init__(self, scenario : Scenario | None = None):
        self.validator = StateValidator()
        self.scenario = scenario

    def _check_goal_state(self, state: StateModel) -> StateModel:
        if self.scenario is None:
            return state
        goal_accomplished = self.scenario.is_in_goal_state(state)
        if (goal_accomplished and not state.goal_accomplished):
            print("Goal accomplished! 😁")
        state.goal_accomplished = goal_accomplished
        return state

    def set_scenario(self, state: StateModel, scenario: Scenario) -> StateModel:
        self.scenario = scenario
        state = scenario.get_initial_state()
        print(f"Goal:\n{scenario.get_goal_state_description()}")
        return state

    def _move_relative(self, state: StateModel, x: int, y: int, claw_index: int) -> StateModel:
        newX = state.claws[claw_index].pos.x + x
        newY = state.claws[claw_index].pos.y + y
        newClawState = replace(state.claws[claw_index], pos=StatePosition(x = newX, y = newY))
        state.claws[claw_index] = newClawState
        #print(f"new position: {newX}, {newY}")
        return state

    def move_horizontally_start(self, state: StateModel, distance: int, claw_index: int) -> StateModel:
        self.validator.move_horizontally(state=state, distance=distance, claw_index=claw_index)
        state.claws[claw_index].moving_horizontally = True
        return self._move_relative(state=state,x=distance, y=0, claw_index=claw_index)
    
    def move_horizontally_end(self, state: StateModel, claw_index: int) -> StateModel:
        state.claws[claw_index].moving_horizontally = False
        return state

    def move_vertically_start(self, state: StateModel, distance: int, claw_index: int) -> StateModel:
        self.validator.move_vertically(state=state, distance=distance, claw_index=claw_index)
        state.claws[claw_index].moving_vertically = True
        return self._move_relative(state=state, x=0, y=distance, claw_index=claw_index)

    def move_vertically_end(self, state: StateModel, claw_index: int) -> StateModel:
        state.claws[claw_index].moving_vertically = False
        return state

    def open_claw_start(self, state: StateModel, claw_index: int) -> StateModel:
        self.validator.open_claw(state, claw_index=claw_index)
        state.claws[claw_index].operating_claw = True
        state.claws[claw_index].open = True
        #print(f"opening claw")
        
        ball_in_claw = state.claws[claw_index].ball
        if ball_in_claw is None:
            return state
        
        print(f"dropping at {state.claws[claw_index].pos}")
        newBall = replace(ball_in_claw, pos = state.claws[claw_index].pos)
        state.claws[claw_index].ball = None
        state.balls.append(newBall)
        return self._check_goal_state(state)

    def close_claw_start(self, state: StateModel, claw_index: int) -> StateModel:
        self.validator.close_claw(state, claw_index=claw_index)
        state.claws[claw_index].operating_claw = True
        state.claws[claw_index].open = False
        #print(f"closing claw")
        ball_to_grab = get_ball_at_current_pos(state, claw_index=claw_index)
        if not ball_to_grab:
            return state
        print(f"grabbing {ball_to_grab} at {state.claws[claw_index].pos}")
        state.claws[claw_index].ball = ball_to_grab
        #remove ball from list
        state.balls = [ball for ball in state.balls if ball.pos != ball_to_grab.pos]
        return self._check_goal_state(state)

    def open_claw_end(self, state: StateModel, claw_index: int) -> tuple[StateModel, bool]:
        droppedBall = next((ball for ball in state.balls if ball.pos == state.claws[claw_index].pos), None)
        
        newState = state
        newState.claws[claw_index].operating_claw = False

        if not droppedBall or not self.scenario:
            return newState, False
        
        return self.scenario.on_ball_dropped(state, droppedBall)

    def close_claw_end(self, state: StateModel, claw_index: int) -> StateModel:
        state.claws[claw_index].operating_claw = False
        return state



from dataclasses import dataclass
from state_update_model import StateBall, StateModel

@dataclass
class ScenarioProgress(object):
    """Represents how many are completed out of a total number of steps."""
    completed: int
    total: int

class Scenario(object):
    """Interface for a specific scenario"""

    _seed: int | None = None

    def __init__(self, seed:int | None = None):
        self._seed = seed

    def get_initial_state(self) -> StateModel:
        """Returns the initial state for the scenario."""
        raise NotImplementedError
    
    def get_progress(self, state: StateModel) -> ScenarioProgress:
        """Returns progress toward goal state."""
        raise NotImplementedError

    def is_in_goal_state(self, state: StateModel) -> bool:
        """Returns true only if state fulfills the goal state criteria."""
        raise NotImplementedError

    def get_goal_state_description(self) -> str:
        """Returns a natural language specification of goal state."""
        raise NotImplementedError
    
    def get_dimensions_description(self) -> str:
        init_state = self.get_initial_state()
        return f"maxX={init_state.max_x}, maxY={init_state.max_y}"

    def on_ball_dropped(self, state: StateModel, ball: StateBall) -> tuple[StateModel, bool]:
        """Overridable state processing after a ball has been dropped."""
        return state, False


from dataclasses import dataclass
from ball_control import IllegalBallControlStateError
from state_utils import (
    get_top_occupied_index,
    get_top_vacant_index,
    is_ball_at_current_pos,
    is_ball_in_claw,
)
from state_update_model import (
    StateModel,
    MIN_X,
    MIN_Y,
)


@dataclass
class StateValidator:
    """Validates operations"""

    def _check_claw_index(self, state: StateModel, claw_index:int):
        if not claw_index < len(state.claws):
            raise IndexError("Claw index out of bounds")

    def _check_claw_collision(self, state: StateModel, claw_index: int, x: int):
        if len(state.claws) > claw_index+1:
            compare_index = claw_index+1
            compare_x = state.claws[compare_index].pos.x
            #print(f"horizontal position of claw {claw_index} ({x}) must be < horizontal position of claw {compare_index} ({compare_x})")
            if x >= compare_x:
                raise IllegalBallControlStateError(f"horizontal position of claw {claw_index} ({x}) must be < horizontal position of claw {compare_index} ({compare_x})")

        if claw_index > 0:
            compare_index = claw_index-1
            compare_x = state.claws[compare_index].pos.x
            #print(f"horizontal position of claw {claw_index} ({x}) must be > horizontal position of claw {compare_index} ({compare_x})")
            if x <= compare_x:
                raise IllegalBallControlStateError(f"horizontal position of claw {claw_index} ({x}) must be > horizontal position of claw {compare_index} ({compare_x})")

    def move_horizontally(self, state: StateModel, distance: int, claw_index: int):
        self._check_claw_index(state=state, claw_index=claw_index)
        
        claw = state.claws[claw_index]

        if (claw.moving_horizontally):
            raise IllegalBallControlStateError("Already moving horizontally")
        
        newX = claw.pos.x + distance
        min_x = max(MIN_X, claw.min_x)
        max_x = min(state.max_x, claw.max_x)
        if newX < min_x or newX > max_x:
            raise IllegalBallControlStateError(f"X coordinate out of bounds x={newX} minX={min_x} maxX={max_x}")
        self._check_claw_collision(state=state, claw_index=claw_index, x=newX)
    
    def move_vertically(self, state: StateModel, distance: int, claw_index: int) -> None:
        self._check_claw_index(state=state, claw_index=claw_index)

        if (state.claws[claw_index].moving_vertically):
            raise IllegalBallControlStateError("Already moving vertically")
        
        newY = state.claws[claw_index].pos.y + distance
        if newY < MIN_Y or newY > state.max_y:
            raise IllegalBallControlStateError(f"Y coordinate out of bounds y={newY} minY={MIN_Y} maxY={state.max_y}")

    def _check_claw_for_ongoing_operations(self, state: StateModel, claw_index: int):
        if (state.claws[claw_index].operating_claw):
            raise IllegalBallControlStateError("Claw already opening or closing")
        
        if (state.claws[claw_index].moving_horizontally or state.claws[claw_index].moving_vertically):
            raise IllegalBallControlStateError("Marble dropped while claw is in motion")

    def open_claw(self, state: StateModel, claw_index: int):
        self._check_claw_index(state=state, claw_index=claw_index)

        if not is_ball_in_claw(state, claw_index=claw_index):
            return
        
        self._check_claw_for_ongoing_operations(state=state, claw_index=claw_index)

        if state.claws[claw_index].pos.y != get_top_vacant_index(state, claw_index=claw_index):
            raise IllegalBallControlStateError(
                f"Illegal drop location. Must be topmost vacant position ({get_top_vacant_index(state, claw_index=claw_index)}). Y={state.claws[claw_index].pos.y}."
            )

    def close_claw(self, state: StateModel, claw_index: int):
        self._check_claw_index(state=state, claw_index=claw_index)

        if not is_ball_at_current_pos(state, claw_index=claw_index):
            return
        
        self._check_claw_for_ongoing_operations(state=state, claw_index=claw_index)

        if state.claws[claw_index].pos.y != get_top_occupied_index(state, claw_index=claw_index):
            raise IllegalBallControlStateError(
                f"Illegal grab. Must be topmost marble position ({get_top_occupied_index(state, claw_index=claw_index)}). Y={state.claws[claw_index].pos.y}."
            )



from scenario import Scenario
#from state_update_model import StateBall

class ScenarioControl(object):
    """Interface for populating a grid"""

    async def set_scenario(self, scenario: Scenario):
        pass


from dataclasses import dataclass, replace
from scenario import Scenario
from state_update_model import (
    MIN_X,
    Highlight,
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch2Scenario(Scenario):
    """Challenge Implementation"""

    def get_goal_state_description(self) -> str:
        return f"Sort balls by weight in leftmost column. Heaviest at the bottom.\n{self.get_dimensions_description()}"
    
    def get_initial_state(self) -> StateModel:
        balls = [
            StateBall(pos=StatePosition(x=1, y=2), color="yellow"),
            StateBall(pos=StatePosition(x=1, y=3), color="blue"),
            StateBall(pos=StatePosition(x=1, y=4), color="green"),
        ]
        highlights = [Highlight(xMin=2, xMax=3, yMin=0, yMax=4, color="lightblue")]
        return replace(get_default_state(), balls = balls, highlights=highlights, max_x=3, max_y=4)

    def is_in_goal_state(self, state: StateModel) -> bool:

        # No ball in claw
        if state.claws[0].ball:
            return False
        
        # yellow at bottom
        if not next((True for y_ball in state.balls if y_ball.color == "yellow" and y_ball.pos == StatePosition(x=MIN_X, y=state.max_y)), False):
            return False

        # green in the middle
        if not next((True for g_ball in state.balls if g_ball.color == "green" and g_ball.pos == StatePosition(x=MIN_X, y=state.max_y-1)), False):
            return False
        
        # blue on top
        if not next((True for b_ball in state.balls if b_ball.color == "blue" and b_ball.pos == StatePosition(x=MIN_X, y=state.max_y-2)), False):
            return False
        
        return True
    
    


from dataclasses import dataclass, replace
from state_update_model import StateBall, StateModel, StatePosition

@dataclass
class RevealColorValueAction:
    """Reveal the values of all balls with the same color when a ball is dropped in a specific location."""

    pos: StatePosition

    def __reveal_ball(self, src_ball: StateBall, target_ball: StateBall | None) -> StateBall | None:
        if target_ball is None:
            return None
        intermediate_ball = replace(target_ball, value_visible = True if target_ball.color==src_ball.color else target_ball.value_visible)
        ret = replace(intermediate_ball, label = f"{intermediate_ball.value}" if intermediate_ball.value_visible else intermediate_ball.label)
        return ret

    def on_ball_dropped(self, state: StateModel, ball: StateBall) -> tuple[StateModel, bool]:
        if ball.pos != self.pos:
            return state, False
        
        balls = [self.__reveal_ball(src_ball=ball, target_ball=b) for b in state.balls]
        claws = [replace(claw, ball=self.__reveal_ball(src_ball=ball, target_ball=claw.ball)) for claw in state.claws]

        return replace(state, balls=balls, claws=claws), True



from dataclasses import dataclass, replace
import random
from reveal_color_value_action import RevealColorValueAction
from scenario import Scenario
from state_update_model import (
    Highlight,
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch12Scenario(Scenario):
    """Challenge Implementation"""

    max_x = 8
    max_y = 7
    colors=["lightblue", "pink", "lightgreen"]
    nof_colors = len(colors)
    reveal_action = RevealColorValueAction(pos=StatePosition(x=max_x, y=max_y))

    def __init__(self, seed: int | None = None):
        super().__init__(seed=seed)

    def get_goal_state_description(self) -> str:
        return f"""
        Two marbles in each of columns 1,2,3. Each marble shall have a value matching the column index.
        Initially you do not know the marble values, only their colors. Each color is associated with an initially unknown random integer value in the range [1, 3].
        A marble must be dropped on position ({self.max_x}, {self.max_y}) to reveal its value, and thereby the value of all marbles of the same color.
        Claw 0 can operate in columns 0-3. Claw 1 can operate in columns 4-{self.max_x}.
        {self.get_dimensions_description()}"""

    def get_initial_state(self) -> StateModel:
        random.seed(self._seed)
    
        nof_balls = 2 * self.nof_colors        
        values = random.sample(range(1, self.nof_colors+1), self.nof_colors)
        min_y = self.max_y+1-nof_balls
        left_y = random.sample(range(min_y, self.max_y+1), nof_balls)
        right_y = random.sample(range(min_y, self.max_y+1), nof_balls)

        left_balls = [
            StateBall(
                pos=StatePosition(x=0, y=left_y[i]),
                color=self.colors[i%self.nof_colors],
                value=values[i%self.nof_colors],
                label="?",
                value_visible=False,
            )
            for i in range(nof_balls)
        ]

        right_balls = [
            StateBall(
                pos=StatePosition(x=self.max_x-1, y=right_y[i]),
                color=self.colors[i%self.nof_colors],
                value=values[i%self.nof_colors],
                label="?",
                value_visible=False,
            )
            for i in range(nof_balls)
        ]

        highlights = [
            Highlight(
                xMin=self.max_x,
                xMax=self.max_x,
                yMin=self.max_y,
                yMax=self.max_y,
                color="lightyellow",
            )
        ]

        claw0 = replace(get_default_state().claws[0], max_x = 3)
        claw1 = replace(claw0, pos=StatePosition(x=6, y=0), min_x=4, max_x = self.max_x)

        return replace(
            get_default_state(),
            balls=left_balls + right_balls,
            max_x=self.max_x,
            max_y=self.max_y,
            highlights=highlights,
            claws=[claw0, claw1]
        )
    
    def is_in_goal_state(self, state: StateModel) -> bool:
        
        # No ball in claw
        if state.claws[0].ball:
            return False
        
        return next((False for ball in state.balls if ball.pos.x < 4 and ball.pos.x != ball.value), True)

    def on_ball_dropped(
        self, state: StateModel, ball: StateBall
    ) -> tuple[StateModel, bool]:
        """Override"""
        return self.reveal_action.on_ball_dropped(state=state, ball=ball)
